#!/usr/bin/env python
"""pollstats - vmstat/dstat like CLI for mmstats

pollstats [options] <fields> <files>...
"""
import argparse
import collections
import mmap
import os
import struct
import sys
import time

import pkg_resources

from mmstats import reader


VERSION = pkg_resources.require('mmstats')[0].version
WARN_LEVEL = 1
INFO_LEVEL = 2
DEBUG_LEVEL = 3

ansi = {
    'black': '\033[0;30m',
    'darkred': '\033[0;31m',
    'darkgreen': '\033[0;32m',
    'darkyellow': '\033[0;33m',
    'darkblue': '\033[0;34m',
    'darkmagenta': '\033[0;35m',
    'darkcyan': '\033[0;36m',
    'gray': '\033[0;37m',

    'darkgray': '\033[1;30m',
    'red': '\033[1;31m',
    'green': '\033[1;32m',
    'yellow': '\033[1;33m',
    'blue': '\033[1;34m',
    'magenta': '\033[1;35m',
    'cyan': '\033[1;36m',
    'white': '\033[1;37m',

    'blackbg': '\033[40m',
    'redbg': '\033[41m',
    'greenbg': '\033[42m',
    'yellowbg': '\033[43m',
    'bluebg': '\033[44m',
    'magentabg': '\033[45m',
    'cyanbg': '\033[46m',
    'whitebg': '\033[47m',

    'reset': '\033[0;0m',
    'bold': '\033[1m',
    'reverse': '\033[2m',
    'underline': '\033[4m',

    'clear': '\033[2J',
#   'clearline': '\033[K',
    'clearline': '\033[2K',
#   'save': '\033[s',
#   'restore': '\033[u',
    'save': '\0337',
    'restore': '\0338',
    'linewrap': '\033[7h',
    'nolinewrap': '\033[7l',

    'up': '\033[1A',
    'down': '\033[1B',
    'right': '\033[1C',
    'left': '\033[1D',

    'default': '\033[0;0m',
}



opts = argparse.ArgumentParser()
opts.add_argument('-c', '--count', type=int)
opts.add_argument('-d', '--delay', type=int, default=1)
opts.add_argument('-f', '--filter', nargs='*', default=[])
opts.add_argument('-n', '--headers', default=20, type=int,
        help='print headers every HEADERS lines')
opts.add_argument('-p', '--prefix', default='', help='field prefix')
opts.add_argument('fields', help='Comma seperated list of fields')
opts.add_argument('files', nargs='+')
opts.add_argument('-v', action='append_const', const=1, dest='verbosity',
        default=[], help='verbosity: warnings=1, info=2, debug=3')
opts.add_argument('--version', action='version',
        version='%%(prog)s %s' %  VERSION)


def get_console_size():
    """Returns rows, columns for current console"""
    #FIXME Only tested on linux
    return map(int, os.popen('stty size', 'r').read().split())


def iter_stats(m):
    """Yields label, value pairs for the given mmstats map"""
    # Hop to the beginning
    m.seek(1)
    while m[m.tell()] != '\x00':
        label_sz = struct.unpack('H', m.read(2))[0]
        label = m.read(label_sz)
        type_sz = struct.unpack('H', m.read(2))[0]
        type_ = m.read(type_sz)
        sz = struct.calcsize(type_)
        idx = struct.unpack('B', m.read_byte())[0]
        if idx == reader.UNBUFFERED_FIELD:
            value = struct.unpack(type_, m.read(sz))[0]
        else:
            idx ^= 1 # Flip bit as the stored buffer is the *write* buffer
            buffers = m.read(sz * 2)
            offset = sz * idx
            value = struct.unpack(type_, buffers[offset:sz+offset])[0]
        yield label, value


Mmap = collections.namedtuple('Mmap', ('file', 'mmap'))


class PollStats(object):
    def __init__(self, args):
        self.args = args
        self.verbosity = len(self.args.verbosity)
        self.fields = args.fields.split(',')
        self.prefix = args.prefix
        if args.prefix:
            self.fields = ["%s%s" % (args.prefix, field)
                    for field in self.fields]
        self.last_vals = dict((field, 0) for field in self.fields)
        self.files = {}
        self._mmap_files()

        # By default filter down to just files that have the given fields
        self.key_filters = set(self.fields)
        self.kv_filters = set()
        for filterstr in args.filter:
            if '=' in filterstr:
                # Filter on exact key-value pairs
                self.kv_filters.add(filterstr.split('='))
            else:
                # Filter on presence of keys
                self.key_filters.add(filterstr)

        self._filter_mmaps()
        self.print_headers()

    def _mmap_files(self):
        for fn in set(self.args.files):
            f = open(fn, 'rb')
            m = mmap.mmap(f.fileno(), 0, prot=mmap.ACCESS_READ)
            if m.read_byte() == reader.VERSION_1:
                self.files[fn] = Mmap(f, m)
            else:
                m.close()
                f.close()
                self.warn('Skipping %s - unknown file format' % fn)

    def _filter_mmaps(self):
        for fn, (f, m) in self.files.items():
            # By default remove files that don't match filters
            remove = True
            fields = dict(pair for pair in iter_stats(m))
            for kf in self.key_filters:
                for k in fields.iterkeys():
                    # Key filters match the start of the key name
                    if k.startswith(kf):
                        remove = False
                        break
                if not remove:
                    break

            if not remove:
                # Matched key filter, don't remove and continue on
                continue

            # Didn't remove file, check kv_filters
            for kf, vf in self.kv_filters:
                if kf in fields and fields[kf] == vf:
                    remove = False
                    break

            if remove:
                # fn didn't match key filter or key/value filter, remove
                self.remove_file(fn)

    def remove_file(self, fn):
        """Remove file `fn` from open mmstat files"""
        m = self.files[fn]
        m.file.close()
        m.mmap.close()
        del self.files[fn]

    def dbg(self, msg):
        if self.verbosity >= DEBUG_LEVEL:
            print >> sys.stderr, msg

    def warn(self, msg):
        if self.verbosity >= WARN_LEVEL:
            print >> sys.stderr, msg

    def print_headers(self):
        #_, width = get_console_size()
        #field_width = (width / len(self.fields)) - 1
        width = 20
        print '|'.join(
            ansi['bold'] + 
                f.replace(self.prefix, '', 1)[:width].center(width)
                + ansi['default']
            for f in self.fields
        )

    def read_once(self):
        cur_vals = collections.defaultdict(int)
        for _, m in self.files.itervalues():
            mvals = dict(pair for pair in iter_stats(m))
            for field in self.fields:
                cur_vals[field] += mvals[field]

        print '|'.join("%s%19d%s " % (ansi['yellow'],
                cur_vals[field] - self.last_vals[field], ansi['default'])
            for field in self.fields)
        self.last_vals = cur_vals

    def run(self):
        if self.args.count:
            for _ in xrange(self.args.count):
                self.read_once()
                time.sleep(self.args.delay)
        else:
            lines_since_header = 0
            while 1:
                if lines_since_header == self.args.headers:
                    self.print_headers()
                    lines_since_header = 0
                self.read_once()
                lines_since_header += 1
                time.sleep(self.args.delay)



def main():
    """CLI Entry Point"""
    p = PollStats(opts.parse_args())
    try:
        p.run()
    except KeyboardInterrupt:
        return

if __name__ == '__main__':
    main()



import ctypes
import os
import sys
import time

from . import fields, libgettid, _mmap
from .defaults import DEFAULT_PATH, DEFAULT_FILENAME


class FieldState(object):
    """Holds field state for each Field instance"""

    def __init__(self, field):
        self.field = field


class BaseMmStats(object):
    """Stats models should inherit from this"""

    def __init__(self, path=DEFAULT_PATH, filename=DEFAULT_FILENAME,
            label_prefix=None):
        """\
        Optionally given a filename or label_prefix, create an MmStats instance
        """
        self._removed = False

        # Setup label prefix
        self._label_prefix = '' if label_prefix is None else label_prefix

        self._offset = 1

        # Store state for this instance's fields
        self._fields = {}

        total_size = self._offset
        #FIXME This is the *wrong* way to initialize stat fields
        for cls in self.__class__.__mro__:
            for attrname, attrval in cls.__dict__.items():
                if (attrname not in self._fields
                        and isinstance(attrval, fields.Field)):
                    total_size += self._add_field(attrname, attrval)

        self._fd, self._filename, self._size, self._mm_ptr = _mmap.init_mmap(
            path=path, filename=filename, size=total_size)
        mmap_t = ctypes.c_char * self._size
        self._mmap = mmap_t.from_address(self._mm_ptr)
        ver = ctypes.c_byte.from_address(self._mm_ptr)
        ver.value = 1  # Version number

        # Finally initialize thes stats
        self._init_fields(total_size)

    def _add_field(self, name, field):
        """Given a name and Field instance, add this field and retun size"""
        # Stats need a place to store their per Mmstats instance state
        state = self._fields[name] = FieldState(field)

        # Call field._new to determine size
        return field._new(state, self.label_prefix, name)

    def _init_fields(self, total_size):
        """Once all fields have been added, initialize them in mmap"""

        for state in self._fields.values():
            # 2nd Call field._init to initialize new stat
            self._offset = state.field._init(state, self._mm_ptr, self._offset)

    @property
    def filename(self):
        return self._filename

    @property
    def label_prefix(self):
        return self._label_prefix

    @property
    def size(self):
        return self._size

    def flush(self, async=False):
        """Flush mmapped file to disk"""
        _mmap.msync(self._mm_ptr, self._size, async)

    def remove(self):
        """Close and remove mmap file - No further stats updates will work"""
        if self._removed:
            # Make calling more than once a noop
            return
        _mmap.munmap(self._mm_ptr, self._size)
        self._size = None
        self._mm_ptr = None
        self._mmap = None
        os.close(self._fd)
        try:
            os.remove(self.filename)
        except OSError:
            # Ignore failed file removals
            pass
        # Remove fields to prevent segfaults
        self._fields = {}
        self._removed = True


class MmStats(BaseMmStats):
    """Mmstats default model base class

    Just subclass, add your own fields, and instantiate:

    >>> from mmstats.models import MmStats
    >>> from mmstats.fields import CounterField
    >>> class MyStats(MmStats):
    ...     errors = CounterField()
    ...
    >>> stats = MyStats()
    >>> stats.errors.inc()
    >>> stats.errors.value
    1L
    """
    pid = fields.StaticUIntField(label="sys.pid", value=os.getpid)
    tid = fields.StaticInt64Field(label="sys.tid", value=libgettid.gettid)
    uid = fields.StaticUInt64Field(label="sys.uid", value=os.getuid)
    gid = fields.StaticUInt64Field(label="sys.gid", value=os.getgid)
    python_version = fields.StaticTextField(label="org.python.version",
            value=lambda: sys.version.replace("\n", ""))
    created = fields.StaticDoubleField(label="sys.created", value=time.time)



import ctypes
import os
import tempfile


BUFFER_IDX_TYPE = ctypes.c_byte
SIZE_TYPE = ctypes.c_ushort
WRITE_BUFFER_UNUSED = 255
DEFAULT_PATH = os.getenv('MMSTATS_PATH', tempfile.gettempdir())
DEFAULT_FILENAME = os.getenv('MMSTATS_FILES', 'mmstats-%PID%-%TID%')
DEFAULT_STRING_SIZE = 255



"""mmstats reader implementation"""
from collections import namedtuple
import mmap
import struct


VERSION_1 = '\x01'
UNBUFFERED_FIELD = 255


def reader(fmt):
    struct_fmt = struct.Struct(fmt)
    size = struct.calcsize(fmt)
    unpacker = struct_fmt.unpack

    def wrapper(v):
        return unpacker(v.read(size))[0]
    wrapper.__name__ = 'unpack_' + fmt

    return wrapper


read_ushort = reader('H')
read_ubyte = reader('B')


Stat = namedtuple('Stat', ('label', 'value'))


class InvalidMmStatsVersion(Exception):
    """Unsupported mmstats version"""


class MmStatsReader(object):
    def __init__(self, data):
        """`data` should be a file-like object (mmap or file)"""
        self.data = data
        rawver = self.data.read(1)
        if rawver == VERSION_1:
            self.version = 1
        else:
            raise InvalidMmStatsVersion(repr(rawver))

    @classmethod
    def from_file(cls, fn):
        return cls(open(fn, 'rb'))

    @classmethod
    def from_mmap(cls, fn):
        f = open(fn, 'rb')
        try:
            mmapf = mmap.mmap(f.fileno(), 0, prot=mmap.ACCESS_READ)
        except:
            f.close()
            raise
        return cls(mmapf)

    def __iter__(self):
        d = self.data
        while 1:
            raw_label_sz = d.read(2)
            if (not raw_label_sz or raw_label_sz == '\x00' or
                    raw_label_sz == '\x00\x00'):
                # EOF
                break
            label_sz = struct.unpack('H', raw_label_sz)[0]
            label = d.read(label_sz).decode('utf8', 'ignore')
            type_sz = read_ushort(d)
            type_ = d.read(type_sz)
            sz = struct.calcsize(type_)
            buf_idx = read_ubyte(d)
            if buf_idx == UNBUFFERED_FIELD:
                value = struct.unpack(type_, d.read(sz))[0]
            else:
                # Flip bit as the stored buffer is the *write* buffer
                buf_idx ^= 1
                buffers = d.read(sz * 2)
                offset = sz * buf_idx
                read_buffer = buffers[offset:(offset + sz)]
                value = struct.unpack(type_, read_buffer)[0]
            if isinstance(value, str):
                # Special case strings as they're \x00 padded
                value = value.split('\x00', 1)[0].decode('utf8', 'ignore')
            yield Stat(label, value)

        try:
            d.close()
        except Exception:
            # Don't worry about exceptions closing the file
            pass



import array
import ctypes
import math
import time

from . import defaults


class DuplicateFieldName(Exception):
    """Cannot add 2 fields with the same name to MmStat instances"""


def _create_struct(label, type_, type_signature, buffers=None):
    """Helper to wrap dynamic Structure subclass creation"""
    if isinstance(label, unicode):
        label = label.encode('utf8')

    fields = [
        ('label_sz', defaults.SIZE_TYPE),
        ('label', ctypes.c_char * len(label)),
        ('type_sig_sz', defaults.SIZE_TYPE),
        ('type_signature', ctypes.c_char * len(type_signature)),
        ('write_buffer', ctypes.c_ubyte),
    ]

    if buffers is None:
        fields.append(('value', type_))
    else:
        fields.append(('buffers', (type_ * buffers)))

    return type("%sStruct" % label.title(),
                (ctypes.Structure,),
                {'_fields_': fields, '_pack_': 1}
            )


class Field(object):
    initial = 0

    def __init__(self, label=None):
        self._struct = None  # initialized in _init
        if label:
            self.label = label
        else:
            self.label = None

    def _new(self, state, label_prefix, attrname, buffers=None):
        """Creates new data structure for field in state instance"""
        # Key is used to reference field state on the parent instance
        self.key = attrname

        # Label defaults to attribute name if no label specified
        if self.label is None:
            state.label = label_prefix + attrname
        else:
            state.label = label_prefix + self.label
        state._StructCls = _create_struct(
                state.label, self.buffer_type,
                self.type_signature, buffers)
        state.size = ctypes.sizeof(state._StructCls)
        return state.size

    def _init(self, state, mm_ptr, offset):
        """Initializes value of field's data structure"""
        state._struct = state._StructCls.from_address(mm_ptr + offset)
        state._struct.label_sz = len(state.label)
        state._struct.label = state.label
        state._struct.type_sig_sz = len(self.type_signature)
        state._struct.type_signature = self.type_signature
        state._struct.write_buffer = defaults.WRITE_BUFFER_UNUSED
        state._struct.value = self.initial
        return offset + ctypes.sizeof(state._StructCls)

    @property
    def type_signature(self):
        return self.buffer_type._type_

    def __repr__(self):
        return '%s(label=%r)' % (self.__class__.__name__, self.label)


class NonDataDescriptorMixin(object):
    """Mixin to add single buffered __get__ method"""

    def __get__(self, inst, owner):
        if inst is None:
            return self
        return inst._fields[self.key]._struct.value


class DataDescriptorMixin(object):
    """Mixin to add single buffered __set__ method"""

    def __set__(self, inst, value):
        inst._fields[self.key]._struct.value = value


class BufferedDescriptorMixin(object):
    """\
    Mixin to add double buffered descriptor methods

    Always read/write as double buffering doesn't make sense for readonly
    fields
    """

    def __get__(self, inst, owner):
        if inst is None:
            return self
        state = inst._fields[self.key]
        # Get from the read buffer
        return state._struct.buffers[state._struct.write_buffer ^ 1]

    def __set__(self, inst, value):
        state = inst._fields[self.key]
        # Set the write buffer
        state._struct.buffers[state._struct.write_buffer] = value
        # Swap the write buffer
        state._struct.write_buffer ^= 1


class ReadOnlyField(Field, NonDataDescriptorMixin):
    def __init__(self, label=None, value=None):
        super(ReadOnlyField, self).__init__(label=label)
        self.value = value

    def _init(self, state, mm, offset):
        if self.value is None:
            # Value can't be None
            raise ValueError("value must be set")
        elif callable(self.value):
            # If value is a callable, resolve it now during initialization
            self.value = self.value()

        # Call super to do standard initialization
        new_offset = super(ReadOnlyField, self)._init(state, mm, offset)
        # Set the static field now
        state._struct.value = self.value

        # And return the offset as usual
        return new_offset


class ReadWriteField(Field, NonDataDescriptorMixin, DataDescriptorMixin):
    """Base class for simple writable fields"""


class DoubleBufferedField(Field):
    """Base class for double buffered writable fields"""
    def _new(self, state, label_prefix, attrname):
        return super(DoubleBufferedField, self)._new(
                state, label_prefix, attrname, buffers=2)

    def _init(self, state, mm_ptr, offset):
        state._struct = state._StructCls.from_address(mm_ptr + offset)
        state._struct.label_sz = len(state.label)
        state._struct.label = state.label
        state._struct.type_sig_sz = len(self.type_signature)
        state._struct.type_signature = self.type_signature
        state._struct.write_buffer = 0
        state._struct.buffers = 0, 0
        return offset + ctypes.sizeof(state._StructCls)


class ComplexDoubleBufferedField(DoubleBufferedField):
    """Base Class for fields with complex internal state like Counters

    Set InternalClass in your subclass
    """
    InternalClass = None

    def _init(self, state, mm_ptr, offset):
        offset = super(ComplexDoubleBufferedField, self)._init(
                state, mm_ptr, offset)
        self._init_internal(state)
        return offset

    def _init_internal(self, state):
        if self.InternalClass is None:
            raise NotImplementedError(
                    "Must set %s.InternalClass" % type(self).__name__)
        state.internal = self.InternalClass(state)

    def __get__(self, inst, owner):
        if inst is None:
            return self
        return inst._fields[self.key].internal


class _InternalFieldInterface(object):
    """Base class used by internal field interfaces like counter"""
    def __init__(self, state):
        self._struct = state._struct

    @property
    def value(self):
        return self._struct.buffers[self._struct.write_buffer ^ 1]

    @value.setter
    def value(self, v):
        self._set(v)

    def _set(self, v):
        # Set the write buffer
        self._struct.buffers[self._struct.write_buffer] = v
        # Swap the write buffer
        self._struct.write_buffer ^= 1


class CounterField(ComplexDoubleBufferedField):
    """Counter field supporting an inc() method and value attribute"""
    buffer_type = ctypes.c_uint64
    type_signature = 'Q'

    class InternalClass(_InternalFieldInterface):
        """Internal counter class used by CounterFields"""
        def inc(self, n=1):
            self._set(self.value + n)


class AverageField(ComplexDoubleBufferedField):
    """Average field supporting an add() method and value attribute"""
    buffer_type = ctypes.c_double

    class InternalClass(_InternalFieldInterface):
        """Internal mean class used by AverageFields"""

        def __init__(self, state):
            _InternalFieldInterface.__init__(self, state)

            # To recalculate the mean we need to store the overall count
            self._count = 0
            # Keep the overall total internally
            self._total = 0.0

        def add(self, value):
            """Add a new value to the average"""
            self._count += 1
            self._total += value
            self._set(self._total / self._count)


class _MovingAverageInternal(_InternalFieldInterface):
    def __init__(self, state):
        _InternalFieldInterface.__init__(self, state)

        self._max = state.field.size
        self._window = array.array('d', [0.0] * self._max)
        self._idx = 0
        self._full = False

    def add(self, value):
        """Add a new value to the moving average"""
        self._window[self._idx] = value
        if self._full:
            self._set(math.fsum(self._window) / self._max)
        else:
            # Window isn't full, divide by current index
            self._set(math.fsum(self._window) / (self._idx + 1))

        if self._idx == (self._max - 1):
            # Reset idx
            self._idx = 0
            self._full = True
        else:
            self._idx += 1


class MovingAverageField(ComplexDoubleBufferedField):
    buffer_type = ctypes.c_double
    InternalClass = _MovingAverageInternal

    def __init__(self, size=100, **kwargs):
        super(MovingAverageField, self).__init__(**kwargs)
        self.size = size


class _TimerContext(object):
    """Class to wrap timer state"""
    def __init__(self, timer=time.time):
        self._timer = timer
        self.start = timer()
        self.end = None

    def get_time(self):
        return self._timer()

    @property
    def done(self):
        """True if timer context has stopped"""
        return self.end is not None

    @property
    def elapsed(self):
        """Returns time elapsed in context"""
        if self.done:
            return self.end - self.start
        else:
            return self.get_time() - self.start

    def stop(self):
        self.end = self.get_time()


class TimerField(MovingAverageField):
    """Moving average field that provides a context manager for easy timings

    As a context manager:
    >>> class T(MmStats):
    ...     timer = TimerField()
    >>> t = T()
    >>> with t.timer as ctx:
    ...     assert ctx.elapsed > 0.0
    >>> assert t.timer.value > 0.0
    >>> assert t.timer.last > 0.0
    """
    def __init__(self, timer=time.time, **kwargs):
        super(TimerField, self).__init__(**kwargs)
        self.timer = timer

    class InternalClass(_MovingAverageInternal):
        def __init__(self, state):
            _MovingAverageInternal.__init__(self, state)
            self._ctx = None
            self.timer = state.field.timer

        def start(self):
            """Start the timer"""
            self._ctx = _TimerContext(self.timer)

        def stop(self):
            """Stop the timer"""
            self._ctx.stop()
            self.add(self._ctx.elapsed)

        def __enter__(self):
            self.start()
            return self._ctx

        def __exit__(self, exc_type, exc_value, exc_tb):
            self.stop()

        @property
        def last(self):
            """Get the last recorded value"""
            if self._ctx is None:
                return 0.0
            else:
                return self._ctx.elapsed


class BufferedDescriptorField(DoubleBufferedField, BufferedDescriptorMixin):
    """Base class for double buffered descriptor fields"""


class UInt64Field(BufferedDescriptorField):
    """Unbuffered read-only 64bit Unsigned Integer field"""
    buffer_type = ctypes.c_uint64
    type_signature = 'Q'


class UIntField(BufferedDescriptorField):
    """32bit Double Buffered Unsigned Integer field"""
    buffer_type = ctypes.c_uint32
    type_signature = 'I'


class IntField(BufferedDescriptorField):
    """32bit Double Buffered Signed Integer field"""
    buffer_type = ctypes.c_int32
    type_signature = 'i'


class ShortField(BufferedDescriptorField):
    """16bit Double Buffered Signed Integer field"""
    buffer_type = ctypes.c_int16


class UShortField(BufferedDescriptorField):
    """16bit Double Buffered Unsigned Integer field"""
    buffer_type = ctypes.c_uint16


class ByteField(ReadWriteField):
    """8bit Signed Integer Field"""
    buffer_type = ctypes.c_byte


class FloatField(BufferedDescriptorField):
    """32bit Float Field"""
    buffer_type = ctypes.c_float


class StaticFloatField(ReadOnlyField):
    """Unbuffered read-only 32bit Float field"""
    buffer_type = ctypes.c_float


class DoubleField(BufferedDescriptorField):
    """64bit Double Precision Float Field"""
    buffer_type = ctypes.c_double


class StaticDoubleField(ReadOnlyField):
    """Unbuffered read-only 64bit Float field"""
    buffer_type = ctypes.c_double


class BoolField(ReadWriteField):
    """Boolean Field"""
    # Avoid potential ambiguity and marshal bools to 0/1 manually
    buffer_type = ctypes.c_byte
    type_signature = '?'

    def __init__(self, initial=False, **kwargs):
        self.initial = initial
        super(BoolField, self).__init__(**kwargs)

    def __get__(self, inst, owner):
        if inst is None:
            return self
        return inst._fields[self.key]._struct.value == 1

    def __set__(self, inst, value):
        inst._fields[self.key]._struct.value = 1 if value else 0


class StringField(ReadWriteField):
    """UTF-8 String Field"""
    initial = ''

    def __init__(self, size=defaults.DEFAULT_STRING_SIZE, **kwargs):
        self.size = size
        self.buffer_type = ctypes.c_char * size
        super(StringField, self).__init__(**kwargs)

    @property
    def type_signature(self):
        return '%ds' % self.size

    def __get__(self, inst, owner):
        if inst is None:
            return self
        return inst._fields[self.key]._struct.value.decode('utf8')

    def __set__(self, inst, value):
        if isinstance(value, unicode):
            value = value.encode('utf8')
            if len(value) > self.size:
                # Round trip utf8 trimmed strings to make sure it's stores
                # valid utf8 bytes
                value = value[:self.size]
                value = value.decode('utf8', 'ignore').encode('utf8')
        elif len(value) > self.size:
            value = value[:self.size]
        inst._fields[self.key]._struct.value = value


class StaticUIntField(ReadOnlyField):
    """Unbuffered read-only 32bit Unsigned Integer field"""
    buffer_type = ctypes.c_uint32
    type_signature = 'I'


class StaticInt64Field(ReadOnlyField):
    """Unbuffered read-only 64bit Signed Integer field"""
    buffer_type = ctypes.c_int64
    type_signature = 'q'


class StaticUInt64Field(ReadOnlyField):
    """Unbuffered read-only 64bit Unsigned Integer field"""
    buffer_type = ctypes.c_uint64
    type_signature = 'Q'


class StaticTextField(ReadOnlyField):
    """Unbuffered read-only UTF-8 encoded String field"""
    initial = ''
    buffer_type = ctypes.c_char * 256
    type_signature = '256s'



import errno
import os
import sys


from mmstats import reader as mmstats_reader


def clean(files):
    alive, removed = 0, 0
    for fn in files:
        sys.stdout.flush()
        if os.path.isdir(fn):
            continue

        try:
            reader = mmstats_reader.MmStatsReader.from_file(fn)
        except mmstats_reader.InvalidMmStatsVersion:
            print 'Invalid file: %s' % fn
            continue
        except IOError as e:
            if e.errno == errno.EACCES:
                print 'Permission denied: %s' % fn
            # Other IOErrors aren't even worth mentioning
            continue

        pid = None
        for k, v in reader:
            if k.endswith('sys.pid'):
                pid = v

        if pid is None:
            print 'File has no sys.pid entry: %s' % fn
            continue

        try:
            os.kill(pid, 0)
        except OSError as e:
            if e.errno == errno.EPERM:
                print ('PID %d is alive but owned by another user, skipping.'
                        % pid)
                continue
            elif e.errno == errno.ESRCH:
                # 'No such process' means we can safely delete this stale pid
                print 'PID %d not found. Deleting %s' % (pid, fn)
            else:
                # Don't assume it's safe to continue after other OSErrors
                raise
        else:
            # PID is alive and well, leave it alone
            alive += 1
            continue

        try:
            os.remove(fn)
            removed += 1
        except OSError as e:
            print('Could not remove %s: %s' % (fn, str(e)))
    print 'Removed %d  -  %d alive' % (removed, alive)


def cli():
    if len(sys.argv) == 1:
        print 'usage: %s MMSTAT_FILES' % os.path.basename(sys.argv[0])
    else:
        clean(sys.argv[1:])


if __name__ == '__main__':
    cli()



import ctypes
import ctypes.util
import errno
libc = ctypes.cdll.LoadLibrary(ctypes.util.find_library('c'))
import mmap as stdlib_mmap
import os


from . import libgettid
from .defaults import DEFAULT_PATH, DEFAULT_FILENAME


PAGESIZE = stdlib_mmap.PAGESIZE


def init_mmap(path=DEFAULT_PATH, filename=DEFAULT_FILENAME, size=PAGESIZE):
    """Given path, filename => filename, size, mmap

    In `filename` "%PID%" and "%TID%" will be replaced with pid and thread id
    """
    # Replace %PID% with actual pid
    filename = filename.replace('%PID%', str(os.getpid()))
    # Replace %TID% with thread id or 0 if thread id is None
    filename = filename.replace('%TID%', str(libgettid.gettid() or 0))

    full_path = os.path.join(path, filename)

    # Create new empty file to back memory map on disk
    fd = os.open(full_path, os.O_CREAT | os.O_TRUNC | os.O_RDWR)
    if size > PAGESIZE:
        if size % PAGESIZE:
            size = size + (PAGESIZE - (size % PAGESIZE))
    else:
        size = PAGESIZE

    # Zero out the file
    os.ftruncate(fd, size)
    m_ptr = mmap(size, fd)
    return (fd, full_path, size, m_ptr)


# Linux consts from /usr/include/bits/mman.h
MS_ASYNC = 1
MS_SYNC = 4


libc.mmap.restype = ctypes.c_void_p
libc.mmap.argtypes = [
    ctypes.c_void_p, # address
    ctypes.c_size_t, # size of mapping
    ctypes.c_int,    # protection
    ctypes.c_int,    # flags
    ctypes.c_int,    # fd
    ctypes.c_int,    # offset (needs to be off_t type?)
]


def mmap(size, fd):
    m_ptr = libc.mmap(None,
                      size,
                      stdlib_mmap.PROT_READ | stdlib_mmap.PROT_WRITE,
                      stdlib_mmap.MAP_SHARED,
                      fd,
                      0
            )
    if m_ptr == -1:
        # Error
        e = ctypes.get_errno()
        raise OSError(e, errno.errorcode[e])
    return m_ptr


libc.msync.restype = ctypes.c_int
libc.msync.argtypes = [
    ctypes.c_void_p, # address
    ctypes.c_size_t, # size of mapping
    ctypes.c_int,    # flags
]


def msync(mm_ptr, size, async=False):
    if async:
        flags = MS_ASYNC
    else:
        flags = MS_SYNC
    status = libc.msync(mm_ptr, size, flags)
    if status == -1:
        e = ctypes.get_errno()
        raise OSError(e, errno.errorcode[e])


libc.munmap.restype = ctypes.c_int
libc.munmap.argtypes = [
    ctypes.c_void_p, # address
    ctypes.c_size_t, # size of mapping
]


def munmap(mm_ptr, size):
    status = libc.munmap(mm_ptr, size)
    if status == -1:
        e = ctypes.get_errno()
        raise OSError(e, errno.errorcode[e])



import sys

from setuptools import setup
from setuptools.extension import Extension

#XXX gettid only works on Linux, don't bother else
if 'linux' in sys.platform:
    exts = [Extension('mmstats._libgettid', sources=['mmstats/_libgettid.c'])]
else:
    exts = []

requirements = ['Flask']


try:
    import argparse
except ImportError:
    # We're probably on Python <2.7, add argparse as a requirement
    requirements.append('argparse')

import mmstats

setup(
    name='mmstats',
    url='https://github.com/schmichael/mmstats',
    version=mmstats.__version__,
    license='APLv2',
    author='Michael Schurter',
    author_email='m@schmichael.com',
    description='Stat, metric, and diagnostic publishing and consuming tools',
    long_description=open('README.rst').read() + '\n\n' +
                     open('CHANGES.rst').read() + '\n\n' +
                     open('TODO.rst').read(),
    packages=['mmstats'],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'mmash=mmstats.mmash:main',
            'slurpstats=mmstats.slurpstats:main',
            'pollstats=mmstats.pollstats:main',
            'cleanstats=mmstats.clean:cli',
        ],
    },
    ext_modules=exts,
    test_suite='tests',
    install_requires=requirements,
    classifiers=['License :: OSI Approved :: Apache Software License'],
    # It might actually be zip-safe, I just hate eggs. File an issue or pull
    # request if mmstats is actually zip_safe and you care
    zip_safe=False
)



import ctypes
import os
import sys
import threading


def _linux_gettid():
    """Return system thread id or pid in single thread process"""
    return _libgettid.gettid()


def _universal_gettid():
    """Give up and just use Python's threading ident"""
    return threading.current_thread().ident


if 'linux' in sys.platform:
    try:
        _PATH = os.path.dirname(os.path.abspath(__file__))
        _libgettid = ctypes.cdll.LoadLibrary(
                os.path.join(_PATH, '_libgettid.so'))
        _libgettid.gettid.restype = ctypes.c_int
        _libgettid.gettid.argtypes = []
    except:
        gettid = _universal_gettid
    else:
        gettid = _linux_gettid
else:
    gettid = _universal_gettid



"""mmash - Flask JSON Web API for publishing mmstats"""
from collections import defaultdict
import glob
import json
import operator
import os
import sys

import flask

from mmstats import reader as mmstats_reader


app = flask.Flask(__name__)
app.config.from_object('mmstats.mmash_settings')
if 'MMASH_SETTINGS' in os.environ:
    app.config.from_envvar('MMASH_SETTINGS')

GLOB = os.path.join(app.config['MMSTATS_DIR'], 'mmstats-*')


def iter_stats():
    """Yields a label at a time from every mmstats file in MMSTATS_DIR"""
    for fn in glob.glob(GLOB):
        try:
            for label, value in mmstats_reader.MmStatsReader.from_mmap(fn):
                yield fn, label, value
        except Exception:
            continue


def find_labels():
    """Returns a set of all available labels"""
    labels = set()
    for fn, label, value in iter_stats():
        labels.add(label)
    return labels


@app.route('/stats/')
def stats():
    return json.dumps(sorted(find_labels()), indent=4)


aggregators = {
    'avg': lambda v: float(sum(v)) / len(v),
    'one': operator.itemgetter(0),
    'max': max,
    'min': min,
    'sum': sum,
}


@app.route('/stats/<statname>')
def getstat(statname):
    stats = defaultdict(list)
    exact = flask.request.args.get('exact')
    for _, label, value in iter_stats():
        if exact and label == statname:
            stats[label].append(value)
        elif label.startswith(statname):
            stats[label].append(value)

    aggr = aggregators.get(flask.request.args.get('aggr'))
    if aggr:
        for label, values in stats.iteritems():
            stats[label] = aggr(values)
    return json.dumps(stats, indent=4)


@app.route('/')
def index():
    return flask.render_template('index.html',
            mmstats_dir=app.config['MMSTATS_DIR'],
            stats=sorted(find_labels()))


def main():
    if len(sys.argv) > 1:
        print __doc__
        print
        print 'Set MMASH_SETTINGS=path/to/settings.py to change settings.'
        return

    app.run(host=app.config['HOST'],
            port=app.config['PORT'],
            debug=app.config['DEBUG']
        )


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import mmap
import os
import sys
import tempfile
import traceback

from mmstats import reader as mmstats_reader


def err(*args):
    """Laziness"""
    sys.stderr.write('%s\n' % ' '.join(map(str, args)))


def slurp_stats(full_fn, m):
    """mmap parsing mainloop"""
    reader = mmstats_reader.MmStatsReader(m)

    print '==>', full_fn
    out = []
    label_max = 0
    for label, value in reader:
        if len(label) > label_max:
            label_max = len(label)
        out.append((label, value))

    for label, value in out:
        if isinstance(value, str):
            value = value.split('\x00', 1)[0]
        print ('  %-' + str(label_max) + 's %s') % (label, value)
    print


def main():
    """MmStats CLI Entry point"""
    # Accept paths and dirs to read as mmstats files from the command line
    stats_files = set()
    for arg in sys.argv[1:]:
        stats_files.add(arg)

    # Only read from tempdir if no files specified on the command line
    if not stats_files:
        tempdir = tempfile.gettempdir()
        stats_files = (os.path.join(tempdir, fn) for fn in os.listdir(tempdir)
                            if fn.startswith('mmstats-'))

    for fn in stats_files:
        with open(fn) as f:
            mmst = mmap.mmap(f.fileno(), 0, prot=mmap.ACCESS_READ)
            try:
                slurp_stats(fn, mmst)
            except Exception:
                err('Error reading: %s' % fn)
                err(traceback.format_exc())
            finally:
                mmst.close()


if __name__ == '__main__':
    main()
